﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace Jegbor
{
    public partial class Jegbor
    {
        /// <summary>
        /// Egy adott cellaintervallumot összegez vagy összeszoroz minden cellaértéket.
        /// </summary>
        /// <param name="acell">A kezdőcella</param>
        /// <param name="bcell">A végcella</param>
        /// <param name="et">A művelet típusa</param>
        /// <returns>Az összeget vagy hányadost</returns>
        public double Interval(string acell, string bcell, EvalType et)
        {
            if ((et != EvalType.Addition) && (et != EvalType.Multiplication))
                throw new ArgumentException("Hiba: Hibás műveletkód.");

            if (!IsCell(acell) || !IsCell(bcell))
                throw new ArgumentException("Hiba: Hibás cellaintervallum.");

            int x1 = 0, x2 = 0, y1, y2;

            char a1, b1;
            a1 = (acell.ToUpper())[0];
            y1 = acell[1] - 49;
            b1 = (bcell.ToUpper())[0];
            y2 = bcell[1] - 49;
            switch (a1)
            {
                case 'A': x1 = 0; break;
                case 'B': x1 = 1; break;
                case 'C': x1 = 2; break;
                case 'D': x1 = 3; break;
                case 'E': x1 = 4; break;
                default: throw new ArgumentException("Hiba: Hibás cella");
            }
            switch (b1)
            {
                case 'A': x2 = 0; break;
                case 'B': x2 = 1; break;
                case 'C': x2 = 2; break;
                case 'D': x2 = 3; break;
                case 'E': x2 = 4; break;
                default: throw new ArgumentException("Hiba: Hibás cella");
            }
            if (x1 > x2)
            {
                int temp = x1;
                x1 = x2;
                x2 = temp;
            }
            if (y1 > y2)
            {
                int temp = y1;
                y1 = y2;
                y2 = temp;
            }

            double result = 0;

            if (et == EvalType.Multiplication)
            {
                result = 1;
            }

            for (int i = y1; i <= y2; i++)
            {
                for (int j = x1; j <= x2; j++)
                {
                    if (et == EvalType.Addition)
                        result += GetCell(j, i).Evaluate(EvalType.Addition);
                    else result *= GetCell(j, i).Evaluate(EvalType.Multiplication);
                }
            }
            

            return 0;
        }

        /// <summary>
        /// Eldönti egy stringről hogy az egy cellát jelöl-e
        /// </summary>
        /// <param name="str">A bemeneti string</param>
        /// <returns>Cella-e a bemeneti string</returns>
        public bool IsCell(string str)
        {
            if (str.Length == 2)
            {
                string t = str.ToUpper();
                char b = t[0];
                int s = str[1] - 48;
                if ((b == 'A') || (b == 'B') || (b == 'C') || (b == 'D') || (b == 'E'))
                {
                    if ((s <= 8) && (s >= 1))
                    {
                        return true;
                    }
                    throw new ArgumentException("Hiba: Hibás cella.");
                }
            }
            return false;
        }

        /// <summary>
        /// Megkeresi a top-level elválasztót az adott függvénytöredékben
        /// </summary>
        /// <param name="fx">A függvénytöredék</param>
        /// <param name="type">Az elválasztó típusa</param>
        /// <returns>Az elválasztó helye (-1 ha nincs)</returns>
        public int FindSeparator(string fx, ref int type)
        {
            int nBr = 0;
            int sId = -1;
            for (int i = 0; i < fx.Length; i++)
            {
                switch (fx[i])
                {
                    case '(':
                        nBr++;
                        break;
                    case ')':
                        nBr--;
                        break;
                    case ';':
                        if (nBr == 0)
                        {
                            sId = i;
                            type = 1;
                        }
                        break;
                    case ':':
                        if (nBr == 0)
                        {
                            sId = i;
                            type = 2;
                        }
                        break;
                    default:
                        break;
                }
                if (sId > -1)
                    break;
            }
            return sId;
        }

        /// <summary>
        /// A táblázat elmentése fájlba
        /// </summary>
        /// <param name="fn">A fájlnév</param>
        public void Save(string fn)
        {
            if (fn.IndexOf(' ') != -1)
            {
                throw new ArgumentException("Hiba: Felesleges szóköz.");
            }
            if (!(fn.Substring(fn.Length - 4, 4) == ".XLH"))
            {
                fn = String.Concat(fn, ".XLH");
            }
            if (File.Exists(fn))
            {
                File.Delete(fn);
            }
            FileStream fs = File.Create(fn);
            StreamWriter sw = new StreamWriter(fs);
            string row;
            for (int i = 0; i <= 7; i++)
            {
                row = "";
                for (int j = 0; j <= 4; j++)
                {
                    row = String.Concat(row, GetCell(j, i).GetText());
                    if (j != 4)
                        row = String.Concat(row, "|");
                }
                sw.WriteLine(row);
            }
            sw.Close();
            fs.Close();
        }

        /// <summary>
        /// Betölti a megadott fájlba elmentett táblázatot.
        /// </summary>
        /// <param name="fn">Fájlnév</param>
        public void Load(string fn)
        {
            if (fn.IndexOf(' ') != -1)
            {
                throw new ArgumentException("Hiba: Felesleges szóköz.");
            }
            if (!(fn.Substring(fn.Length - 4, 4) == ".XLH"))
            {
                fn = String.Concat(fn, ".XLH");
            }
            if (!File.Exists(fn))
            {
                throw new ArgumentException("Hiba: A megadott fájl nem létezik.");
            }
            FileStream fs = File.OpenRead(fn);
            StreamReader sr = new StreamReader(fs);
            for (int i = 0; i <= 7; i++)
            {
                string row = sr.ReadLine();
                string[] val = row.Split('|');
                int nv = val.Length;
                if (nv != 5)
                {
                    throw new ArgumentException("Hiba a fájlban: A(z) "+i+". sorban nem megfelelő a cellák száma.");
                }
                for (int j = 0; j <= 4; j++)
                {
                    SetCell(j, i, val[j]);
                }
            }
        }

        /// <summary>
        /// Függvény értelmezése és elvégzése
        /// </summary>
        /// <param name="fx">A függvény</param>
        /// <param name="et">Művelet típusa (ha megállapítható)</param>
        /// <returns>A végérték</returns>
        public double ParseFunction(string fx, EvalType et)
        {
            string inner, outer;

            if (IsCell(fx))
            {                     
                return GetCell(fx).Evaluate(et);
            }
            else if (fx.IndexOf('(') == -1)
            {
                return Convert.ToDouble(fx);
            }

            if (fx.Length == 3)
            {

            }

            if (fx[fx.Length - 1] != ')')
            {
                throw new ArgumentException("Hiba: Hiányzó zárójel.");
            }
            if (fx.IndexOf(' ') != -1)
            {
                throw new ArgumentException("Hiba: Felesleges szóköz.");
            }
            int outindex = fx.IndexOf('(');
            if (outindex == -1)
            {
                throw new ArgumentException("Hiba: A bemenet nem függvény.");
            }
            outer = fx.Substring(0, outindex);
            inner = fx.Substring(outindex + 1, fx.Length - (outindex + 2));

            string arg1, arg2;
            int septype = 0;
            int sep = FindSeparator(inner,ref septype);
            if (sep == -1)
            {
                throw new ArgumentException("Hiba: Túl kevés paraméter.");
            }
            arg1 = inner.Substring(0, sep);
            arg2 = inner.Substring(sep + 1, inner.Length - (sep + 1));

            outer = outer.ToUpper();
            switch (outer)
            {
                case "OSSZEG":
                    et = EvalType.Addition;
                    break;
                case "SZORZAT":
                    et = EvalType.Multiplication;
                    break;
                case "HANYADOS":
                    et = EvalType.Division;
                    break;
                default:
                    throw new ArgumentException("Hiba: Nem létező függvény.");
            }

            if (septype == 1)
            {
                double val1 = ParseFunction(arg1,et);
                double val2 = ParseFunction(arg2,et);
                switch (et)
                {
                    case EvalType.Addition:
                        return val1 + val2;
                    case EvalType.Multiplication:
                        return val1 * val2;
                    case EvalType.Division:
                        if (val2 == 0)
                        {
                            throw new ArgumentException("Hiba: Nullával való osztás.");
                        }
                        else return val1 / val2;
                }
            }
            else if (septype == 2)
            {
                if (!IsCell(arg1) || !IsCell(arg2))
                {
                    throw new ArgumentException("Hiba: A ':' operátor mindkét oldalán cellának kell állnia!");
                }
                if (et == EvalType.Division)
                {
                    throw new ArgumentException("Hiba: Cellaintervallumon nem végezhető osztás.");
                }
                return Interval(arg1, arg2, et);
            }

            return 0;
        }

        /// <summary>
        /// Egy tizedestört megformázása
        /// </summary>
        /// <param name="x">A tizedestört</param>
        /// <param name="decimals">Tizedesjegyek száma</param>
        /// <returns></returns>
        public static string FormatDouble(double x, int decimals)
        {
            string r = x.ToString();
            int seppos = r.IndexOf('.');
            bool negative = false;
            if (r[0] == '-') negative = true;
            if (seppos == -1)
            {
                if (decimals == 0)
                    return r;
                r += ".";
                for (int i = 1; i <= decimals; i++)
                    r += "0";
            }
            else
            {
                if (decimals == 0)
                {
                    r = r.Substring(0, seppos);
                    return r;
                }
                
                int dec = r.Length - (seppos + 1);

                if (dec < decimals)
                {
                    for (int i = 1; i <= decimals - dec; i++)
                    {
                        r += "0";
                        return r;
                    }
                }

                if (decimals == dec)
                {
                    return r;
                }

                if (decimals < dec)
                {
                    int[] numbers = new int[9];
                    int[] decs = new int[9];
                    int n = 0;
                    numbers[0] = -1;
                    int sp = negative ? 1 : 0;
                    for (int i = 0; i < seppos; i++)
                    {
                        numbers[i+1] = Convert.ToInt32(r[i+sp]);
                        n++;
                    }

                    for (int i = seppos+1; i < r.Length; i++)
                    {
                        decs[i-seppos-1] = r[i];
                    }

                    if (!negative)
                    {
                        if (decs[decimals] < 5)
                        {
                            r = r.Substring(0, seppos + decimals + 1);
                        }
                        else
                        {
                            bool carry = true;
                            for (int i = decimals - 1; i >= 0; i--)
                            {
                                if (!carry) break;
                                decs[i] += 1;
                                if (decs[i] == 10)
                                {
                                    carry = true;
                                    decs[i] = 0;
                                }
                                else carry = false;
                            }
                            if (carry)
                            {
                                for (int i = n; i >= 0; i--)
                                {
                                    if (!carry) break;
                                    numbers[i] += 1;
                                    if (numbers[i] == 10)
                                    {
                                        carry = true;
                                        numbers[i] = 0;
                                    }
                                    else carry = false;
                                }
                            }
                        }
                        string nr = "";
                        for (int i = 0; i <= n; i++)
                        {
                            if (numbers[i] != -1)
                                nr += numbers[i].ToString();
                        }
                        nr += ".";
                        for (int i = 0; i < decimals; i++)
                        {
                            nr += decs[i];
                        }
                        return nr;
                    }
                    else
                    {
                        if (decs[decimals] < 5)
                        {
                            r = r.Substring(0, seppos + decimals + 1);
                        }
                        bool fnz = false;
                        for (int i = 1; i <= n; i++)
                        {
                            if (r[i] != '0')
                                fnz = true;
                        }
                        for (int i = seppos+1; i < r.Length; i++)
                        {
                            if (r[i] != '0')
                                fnz = true;
                        }
                        if (fnz == false)
                        {
                            switch (decimals)
                            {
                                case 1:
                                    return "0.0";
                                case 2:
                                    return "0.00";
                                case 3:
                                    return "0.000";
                                default:
                                    return "0";
                            }
                        }

                        if (decs[decimals] >= 5)
                        {
                            bool carry = true;
                            for (int i = decimals - 1; i >= 0; i--)
                            {
                                if (!carry) break;
                                decs[i] += 1;
                                if (decs[i] == 10)
                                {
                                    carry = true;
                                    decs[i] = 0;
                                }
                                else carry = false;
                            }
                            if (carry)
                            {
                                for (int i = n; i >= 0; i--)
                                {
                                    if (!carry) break;
                                    numbers[i] += 1;
                                    if (numbers[i] == 10)
                                    {
                                        carry = true;
                                        numbers[i] = 0;
                                    }
                                    else carry = false;
                                }
                            }
                        }
                        string nr = "-";
                        for (int i = 0; i <= n; i++)
                        {
                            if (numbers[i] != -1)
                                nr += numbers[i].ToString();
                        }
                        nr += ".";
                        for (int i = 0; i < decimals; i++)
                        {
                            nr += decs[i];
                        }
                        return nr;
                    }
                }
                
            }
            throw new ArgumentException("Hiba: Sikertelen formázási kísérlet.");
        }

    }
}